import codecs
import datetime
import http.client
import json
import urllib.parse
from io import BytesIO

import chardet

from pycurl_requests.exceptions import HTTPError


class Request:
    def __init__(self,
                 method=None,
                 url=None,
                 headers=None,
                 files=None,
                 data=None,
                 params=None,
                 auth=None,
                 cookies=None,
                 hooks=None,
                 json=None):

        self.method = method
        self.url = url
        self.headers = headers
        self.files = files
        self.data = data
        self.json = json
        self.params = params
        self.auth = auth
        self.cookies = cookies
        self.hooks = hooks

    def deregister_hook(self, event, hook):
        raise NotImplementedError

    def prepare(self):
        prepared = PreparedRequest()
        prepared.prepare(
            method=self.method,
            url=self.url,
            headers=self.headers,
            files=self.files,
            data=self.data,
            params=self.params,
            auth=self.auth,
            cookies=self.cookies,
            hooks=self.hooks,
            json=self.json)

        return prepared

    def register_hook(self, event, hook):
        raise NotImplementedError


class Response:
    def __init__(self,
                 prepared_request: Request = None,
                 elapsed: datetime.timedelta = None,
                 status_code: int = None,
                 reason: str = None,
                 headers: http.client.HTTPMessage = None,
                 encoding: str = None,
                 url: str = None,
                 buffer: BytesIO = None):
        self._prepared_request = prepared_request
        self._elapsed = elapsed
        self._status_code = status_code
        self._reason = reason
        self._headers = headers
        self.encoding = encoding
        self._url = url
        self._buffer = buffer

    @property
    def apparent_encoding(self):
        return chardet.detect(self.content)['encoding']

    def close(self):
        # Not implemented
        pass

    @property
    def content(self):
        return self._buffer.getvalue()

    @property
    def cookies(self):
        return NotImplemented

    @property
    def elapsed(self):
        return self._elapsed

    @property
    def headers(self):
        return self._headers

    @property
    def history(self):
        return NotImplemented

    @property
    def is_permanent_redirect(self):
        # Moved Permanently (HTTP 301) or Permanent Redirect (HTTP 308)
        return self.status_code in {301, 308}

    def iter_content(self, chunk_size=1, decode_unicode=False):
        chunk_size = chunk_size or -1
        decoder = codecs.getincrementaldecoder(self.encoding)('replace') if self.encoding and decode_unicode else None
        for chunk in iter(lambda: self._buffer.read1(chunk_size), b''):
            if decoder:
                yield decoder.decode(chunk)
            else:
                yield chunk

        if decoder:
            # Make sure we finalize the decoder (may yield replacement character)
            tail = decoder.decode(b'', True)
            if tail:
                yield tail

    def iter_lines(self, chunk_size=512, decode_unicode=False, delimiter=None):
        delimiter = delimiter or ('\n' if self.encoding and decode_unicode else b'\n')

        leftover = None
        for chunk in self.iter_content(chunk_size, decode_unicode=decode_unicode):
            while True:
                parts = chunk.split(delimiter, 1)
                if len(parts) == 1:
                    leftover = parts[0]
                    break
                elif leftover:
                    chunk = parts[1]
                    yield leftover + parts[0]
                else:
                    chunk = parts[1]
                    yield parts[0]

        if leftover is not None:
            yield leftover

    def json(self, **kwargs):
        return json.loads(self.content, **kwargs)

    @property
    def links(self):
        return NotImplemented

    @property
    def next(self):
        return NotImplemented

    @property
    def ok(self):
        return self.status_code < 400

    def raise_for_status(self):
        if 400 <= self.status_code < 500:
            raise HTTPError('{s.status_code} Client Error: {s.reason} for url: {s.url}'.format(s=self),
                            response=self)

        if 500 <= self.status_code < 600:
            raise HTTPError('{s.status_code} Client Error: {s.reason} for url: {s.url}'.format(s=self),
                            response=self)

    @property
    def raw(self):
        return self._buffer

    @property
    def reason(self):
        return self._reason

    @property
    def request(self):
        return self._prepared_request

    @property
    def status_code(self):
        return self._status_code

    @property
    def text(self):
        return self.content.decode()

    @property
    def url(self):
        return self._url


class PreparedRequest:
    def __init__(self):
        self.method = None
        self.url = None
        self.headers = None
        self.body = None
        self.hooks = None

    @property
    def path_url(self):
        return urllib.parse.urlsplit(self.url).path

    def prepare(self,
                method=None,
                url=None,
                headers=None,
                files=None,
                data=None,
                params=None,
                auth=None,
                cookies=None,
                hooks=None,
                json=None):

        self.prepare_method(method)
        self.prepare_url(url, params)
        self.prepare_headers(headers)
        self.prepare_cookies(cookies)
        self.prepare_body(data, files, json)
        self.prepare_auth(auth, url)
        self.prepare_hooks(hooks)

    def prepare_method(self, method):
        self.method = method.upper()

    def prepare_url(self, url, params):
        if params:
            # Rebuild the URL with params included
            params = list(params.items()) if isinstance(params, dict) else params
            parts = urllib.parse.urlsplit(url)
            query = urllib.parse.urlencode(urllib.parse.parse_qsl(parts.query) + params, doseq=True)
            self.url = urllib.parse.urlunsplit(parts[:3] + (query,) + parts[4:])
        else:
            self.url = url

    def prepare_headers(self, headers):
        # NOTE: Only user-defined headers, not those set by libcurl
        self.headers = headers or http.client.HTTPMessage()

    def prepare_cookies(self, cookies):
        # FIXME: Not implemented
        pass

    def prepare_content_length(self, body):
        # FIXME: Not implemented
        pass

    def prepare_body(self, data, files, json=None):
        # FIXME: Not implemented
        pass

    def prepare_auth(self, auth, url=''):
        # FIXME: Not implemented
        pass

    def prepare_hooks(self, hooks):
        # FIXME: Not implemented
        pass



class RequestException(IOError):
    def __init__(self, *args, **kwargs):
        self.response = kwargs.pop('response', None)
        self.request = kwargs.pop('request', None)

        if self.response is not None and not self.request and hasattr(self.response, 'request'):
            self.request = self.response.request

        super().__init__(*args, **kwargs)


class ConnectionError(RequestException):
    pass


class HTTPError(RequestException):
    pass


class URLRequired(RequestException):
    pass


class TooManyRedirects(RequestException):
    pass


class Timeout(RequestException):
    pass


class ConnectTimeout(ConnectionError, Timeout):
    pass


class ReadTimeout(Timeout):
    pass



import datetime
import http.client
import json
import threading
import urllib.parse
from http.server import HTTPServer, BaseHTTPRequestHandler

import pytest

import pycurl_requests as requests


@pytest.fixture(scope='module')
def http_server():
    httpd = HTTPServer(('127.0.0.1', 0), HTTPRequestHandler)
    httpd.base_url = 'http://{}:{}'.format(*httpd.server_address)
    thread = threading.Thread(target=httpd.serve_forever)
    thread.start()
    try:
        yield httpd
    finally:
        httpd.shutdown()
        thread.join()


class HTTPRequestHandler(BaseHTTPRequestHandler):
    def log_message(self, format, *args):
        # Mute HTTP logging
        pass

    def do_GET(self):
        if self.url.path == '/hello':
            self.response('Hello\nWorld\n')
        elif self.url.path == '/params':
            self.response('\n'.join(('{}: {}'.format(n, v)
                                     for n, v in urllib.parse.parse_qsl(self.url.query))))
        elif self.url.path == '/headers':
            self.response('\n'.join(('{}: {}'.format(n, v)
                                     for n, v in self.headers.items())))
        elif self.url.path == '/redirect':
            self.response('Redirecting...\n', (302, 'Found'),
                          headers={'Location': '/redirected'})
        elif self.url.path == '/redirected':
            self.response('Redirected\n')
        elif self.url.path == '/json':
            self.response(json.dumps({'Hello': 'World'}), content_type='application/json')
        else:
            self.send_error(404, 'Not Found')

    @property
    def url(self):
        if not hasattr(self, '_url'):
            # For some annoying reason this can't be set in `__init__`
            self._url = urllib.parse.urlsplit(self.path)

        return self._url

    def response(self, body, status=None, content_type=None, headers=None):
        body = body.encode('utf-8') if isinstance(body, str) else body
        status = status or (200, 'OK')
        content_type = content_type or 'text/html; charset=UTF-8'
        headers = headers or {}
        self.send_response(*status)
        self.send_header('Content-Type', content_type)
        self.send_header('Content-Length', len(body))
        for key, value in headers.items():
            self.send_header(key, value)
        self.end_headers()

        self.wfile.write(body)


def test_get(http_server):
    response = requests.get(http_server.base_url + '/hello')
    assert isinstance(response, requests.Response)
    response.raise_for_status()

    assert isinstance(response.request, requests.PreparedRequest)
    assert response.request.method == 'GET'
    assert response.request.url == http_server.base_url + '/hello'
    assert response.request.path_url == '/hello'
    assert isinstance(response.request.headers, http.client.HTTPMessage)
    assert response.request.body is None

    assert response.elapsed > datetime.timedelta(0)
    assert response.url.endswith('/hello')
    assert response.status_code == 200
    assert response.reason == 'OK'
    assert response.headers['content-type'] == 'text/html; charset=UTF-8'
    assert response.headers['content-length'] == str(len(b'Hello\nWorld\n'))
    assert response.encoding == 'UTF-8'
    assert response.apparent_encoding == 'ascii'
    assert response.content == b'Hello\nWorld\n'
    assert response.text == 'Hello\nWorld\n'


def test_get_notfound(http_server):
    response = requests.get(http_server.base_url + '/not_found')
    with pytest.raises(requests.HTTPError) as err:
        response.raise_for_status()
    assert err.value.response.status_code == 404


def test_get_params(http_server):
    response = requests.get(http_server.base_url + '/params', params={'q': 'foo', 'r': 'bar'})
    response.raise_for_status()

    assert response.text == 'q: foo\nr: bar'


def test_get_headers(http_server):
    response = requests.get(http_server.base_url + '/headers', headers={'Foo': 'foo', 'Bar': 'bar'})
    response.raise_for_status()
    print(response.text)

    headers = http.client.HTTPMessage()
    for line in response.text.splitlines():
        name, value = line.split(':', 1)
        headers[name] = value.strip()

    assert headers['Foo'] == 'foo'
    assert headers['Bar'] == 'bar'


def test_get_redirect_nofollow(http_server):
    response = requests.get(http_server.base_url + '/redirect', allow_redirects=False)
    response.raise_for_status()

    assert response.url.endswith('/redirect')
    assert response.status_code == 302
    assert response.text == 'Redirecting...\n'


def test_get_redirect(http_server):
    response = requests.get(http_server.base_url + '/redirect')
    response.raise_for_status()

    assert response.url.endswith('/redirected')
    assert response.text == 'Redirected\n'


def test_get_json(http_server):
    response = requests.get(http_server.base_url + '/json')
    response.raise_for_status()

    assert response.json() == {'Hello': 'World'}


def test_get_iter_content(http_server):
    response = requests.get(http_server.base_url + '/hello')
    response.raise_for_status()

    assert response.content == b'Hello\nWorld\n'
    it = response.iter_content(5, decode_unicode=True)
    assert next(it) == 'Hello'
    assert next(it) == '\nWorl'
    assert next(it) == 'd\n'
    with pytest.raises(StopIteration):
        assert next(it)


@pytest.mark.parametrize('delimiter', [None, '\n'])
def test_get_iter_lines(delimiter, http_server):
    response = requests.get(http_server.base_url + '/hello')
    response.raise_for_status()

    assert response.content == b'Hello\nWorld\n'
    it = response.iter_lines(5, decode_unicode=True, delimiter=delimiter)
    assert next(it) == 'Hello'
    assert next(it) == 'World'
    assert next(it) == ''
    with pytest.raises(StopIteration):
        assert next(it)



import contextlib
import datetime
import http.client
from io import BytesIO
from typing import *

import pycurl

from pycurl_requests import models


def request(method, url, *, curl=None, allow_redirects=True, **kwargs):
    if method != 'GET':
        raise NotImplementedError

    curl = curl or pycurl.Curl()

    request = models.Request(method, url, **kwargs)
    prepared = request.prepare()

    buffer = BytesIO()
    reason = None
    headers = http.client.HTTPMessage()

    def header_function(line):
        nonlocal reason

        try:
            # Some servers return UTF-8 status
            line = line.decode('utf-8')
        except UnicodeDecodeError:
            # Fall back to latin-1
            line = line.decode('iso-8859-1')

        if reason is None:
            _, _, reason = line.split(' ', 2)
            reason = reason.strip()
            return

        if ':' not in line:
            return

        name, value = line.split(':', 1)
        headers[name] = value.strip()

    start_time = datetime.datetime.now(tz=datetime.timezone.utc)

    with contextlib.closing(curl) as c:
        c.setopt(c.URL, prepared.url)
        c.setopt(c.HTTPHEADER, ['{}: {}'.format(n, v) for n, v in prepared.headers.items()])
        c.setopt(c.HEADERFUNCTION, header_function)
        c.setopt(c.WRITEDATA, buffer)
        if allow_redirects:
            c.setopt(c.FOLLOWLOCATION, 1)

        c.perform()

        status_code = c.getinfo(c.RESPONSE_CODE)
        effective_url = c.getinfo(c.EFFECTIVE_URL)

    mime_type, params = parse_content_type(headers['Content-Type'])
    encoding = params.get('charset')

    buffer.seek(0)

    end_time = datetime.datetime.now(tz=datetime.timezone.utc)

    return models.Response(
        prepared_request=prepared,
        elapsed=end_time - start_time,
        status_code=status_code,
        reason=reason,
        headers=headers,
        encoding=encoding,
        url=effective_url,
        buffer=buffer,
    )


def parse_content_type(content_type: str) -> (str, Dict[str, str]):
    """Parse value of `Content-Type` header"""
    parameters = content_type.split(';')
    mime_type = parameters[0].strip()
    params = http.client.HTTPMessage()
    for parameter in parameters[1:]:
        name, value = parameter.split('=', 1)
        name, value = name.strip(), value.strip()
        params[name] = value

    return mime_type, params






from pycurl_requests import _pycurl


def request(method, url, **kwargs):
    return _pycurl.request(method, url, **kwargs)


def head(url, **kwargs):
    return request('HEAD', url, **kwargs)


def get(url, params=None, **kwargs):
    return request('GET', url, params=params, **kwargs)


def post(url, data=None, json=None, **kwargs):
    return request('POST', url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    return request('PUT', url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    return request('PATCH', url, data=data, **kwargs)


def delete(url, params=None, **kwargs):
    return request('GET', url, params=params, **kwargs)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='pycurl-requests',
    version='0.0.1',
    description='A Requests-compatible interface for pycURL',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/dcoles/pycurl-requests',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=[
        'pycurl',
        'chardet',
        'pytest',
    ],
)



from .api import request, get, head, post, patch, put, delete
from .exceptions import RequestException, ConnectionError, HTTPError, URLRequired, TooManyRedirects, ConnectTimeout, ReadTimeout, Timeout
from .models import Request, Response, PreparedRequest
